# Storage configuration spoke classes
#
# Copyright (C) 2011-2014  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#                    Chris Lumens <clumens@redhat.com>
#

"""
    TODO:

        - add button within sw_needs text in options dialogs 2,3
        - udev data gathering
            - udev fwraid, mpath would sure be nice
        - status/completed
            - what are noteworthy status events?
                - disks selected
                    - exclusiveDisks non-empty
                - sufficient space for software selection
                - autopart selected
                - custom selected
                    - performing custom configuration
                - storage configuration complete
        - spacing and border width always 6

"""

from gi.repository import Gdk, GLib, AnacondaWidgets, Gtk, Pango

from pyanaconda.ui.communication import hubQ
from pyanaconda.ui.lib.disks import getDisks, isLocalDisk, applyDiskSelection, checkDiskSelection
from pyanaconda.ui.gui import GUIObject
from pyanaconda.ui.lib.disks import getDiskDescription
from pyanaconda.ui.gui.spokes import NormalSpoke
from pyanaconda.ui.gui.spokes.lib.cart import SelectedDisksDialog
from pyanaconda.ui.gui.spokes.lib.passphrase import PassphraseDialog
from pyanaconda.ui.gui.spokes.lib.detailederror import DetailedErrorDialog
from pyanaconda.ui.gui.spokes.lib.resize import ResizeDialog
from pyanaconda.ui.gui.spokes.lib.dasdfmt import DasdFormatDialog
from pyanaconda.ui.gui.spokes.lib.refresh import RefreshDialog
from pyanaconda.ui.categories.system import SystemCategory
from pyanaconda.ui.gui.utils import escape_markup, gtk_action_nowait, ignoreEscape, really_hide
from pyanaconda.ui.helpers import StorageChecker
from pyanaconda.storage_utils import on_disk_storage

from pyanaconda.kickstart import doKickstartStorage, refreshAutoSwapSize, resetCustomStorageData
from blivet import arch
from blivet.size import Size
from blivet.devices import MultipathDevice, ZFCPDiskDevice, iScsiDiskDevice, MultipathDevice
from blivet.errors import StorageError, DasdFormatError
from blivet.platform import platform
from blivet.devicelibs import swap as swap_lib
from blivet.devicelibs.dasd import make_unformatted_dasd_list, format_dasd, is_ldl_dasd
from pyanaconda.threads import threadMgr, AnacondaThread
from pyanaconda.product import productName
from pyanaconda.flags import flags
from pyanaconda.i18n import _, C_, CN_, P_
from pyanaconda import constants, iutil, isys
from pyanaconda.bootloader import BootLoaderError
from pyanaconda.virtuozzo_c import DoVZAutopartPreselect, minimal_hardware

from pykickstart.constants import CLEARPART_TYPE_NONE, CLEARPART_TYPE_ALL, AUTOPART_TYPE_LVM
from pykickstart.errors import KickstartValueError

import sys

import logging
log = logging.getLogger("anaconda")

__all__ = ["StorageSpoke"]

# Response ID codes for all the various buttons on all the dialogs.
RESPONSE_CANCEL = 0
RESPONSE_MODIFY_SW = 2
RESPONSE_RECLAIM = 3
RESPONSE_QUIT = 4

# Columns in the table of disks
COL_DISK = 0
COL_SPACE = 1
COL_TYPE = 2
COL_SIZE = 3
COL_BOOT = 4
COL_SYSTEM = 5
COL_DATASTORE = 6
COL_CACHE = 7

class InstallOptionsDialogBase(GUIObject):
    uiFile = "spokes/storage.glade"

    def __init__(self, *args, **kwargs):
        self.payload = kwargs.pop("payload", None)
        GUIObject.__init__(self, *args, **kwargs)

        self._grabObjects()

    def _grabObjects(self):
        pass

    def run(self):
        rc = self.window.run()
        self.window.destroy()
        return rc

    def _modify_sw_link_clicked(self, label, uri):
        if self._software_is_ready():
            self.window.response(RESPONSE_MODIFY_SW)

        return True

    def _get_sw_needs_text(self, required_space, auto_swap):
        tooltip = _("Please wait... software metadata still loading.")
        sw_text = (_("Your current <a href=\"\" title=\"%(tooltip)s\"><b>%(product)s</b> software "
                     "selection</a> requires <b>%(total)s</b> of available "
                     "space, including <b>%(software)s</b> for software and "
                     "<b>%(swap)s</b> for swap space.")
                   % {"tooltip": escape_markup(tooltip),
                      "product": escape_markup(productName),
                      "total": escape_markup(str(required_space + auto_swap)),
                      "software": escape_markup(str(required_space)),
                      "swap": escape_markup(str(auto_swap))})
        return sw_text

    # Methods to handle sensitivity of the modify button.
    def _software_is_ready(self):
        # FIXME:  Would be nicer to just ask the spoke if it's ready.
        return (not threadMgr.get(constants.THREAD_PAYLOAD) and
                not threadMgr.get(constants.THREAD_SOFTWARE_WATCHER) and
                not threadMgr.get(constants.THREAD_CHECK_SOFTWARE) and
                self.payload.baseRepo is not None)

    def _check_for_storage_thread(self, button):
        if self._software_is_ready():
            button.set_has_tooltip(False)

            # False means this function should never be called again.
            return False
        else:
            return True

    def _add_modify_watcher(self, widget):
        # If the payload fetching thread is still running, the user can't go to
        # modify the software selection screen.  Thus, we have to set the button
        # insensitive and wait until software selection is ready to go.
        if not self._software_is_ready():
            GLib.timeout_add_seconds(1, self._check_for_storage_thread, widget)

class NeedSpaceDialog(InstallOptionsDialogBase):
    builderObjects = ["need_space_dialog"]
    mainWidgetName = "need_space_dialog"

    def _grabObjects(self):
        self.disk_free_label = self.builder.get_object("need_space_disk_free_label")
        self.fs_free_label = self.builder.get_object("need_space_fs_free_label")

    def _set_free_space_labels(self, disk_free, fs_free):
        self.disk_free_label.set_text(str(disk_free))
        self.fs_free_label.set_text(str(fs_free))

    # pylint: disable=arguments-differ
    def refresh(self, required_space, auto_swap, disk_free, fs_free):
        sw_text = self._get_sw_needs_text(required_space, auto_swap)
        label_text = _("%s The disks you've selected have the following "
                       "amounts of free space:") % sw_text
        label = self.builder.get_object("need_space_desc_label")
        label.set_markup(label_text)
        label.connect("activate-link", self._modify_sw_link_clicked)

        self._set_free_space_labels(disk_free, fs_free)

        label_text = _("<b>You don't have enough space available to install "
                       "%s</b>.  You can shrink or remove existing partitions "
                       "via our guided reclaim space tool, or you can adjust your "
                       "partitions on your own in the custom partitioning "
                       "interface.") % escape_markup(productName)
        self.builder.get_object("need_space_options_label").set_markup(label_text)

        self._add_modify_watcher(label)

class NoSpaceDialog(InstallOptionsDialogBase):
    builderObjects = ["no_space_dialog"]
    mainWidgetName = "no_space_dialog"

    def _grabObjects(self):
        self.disk_free_label = self.builder.get_object("no_space_disk_free_label")
        self.fs_free_label = self.builder.get_object("no_space_fs_free_label")

    def _set_free_space_labels(self, disk_free, fs_free):
        self.disk_free_label.set_text(str(disk_free))
        self.fs_free_label.set_text(str(fs_free))

    # pylint: disable=arguments-differ
    def refresh(self, required_space, auto_swap, disk_free, fs_free):
        label_text = self._get_sw_needs_text(required_space, auto_swap)
        label_text += (_("  You don't have enough space available to install "
                         "<b>%(product)s</b>, even if you used all of the free space "
                         "available on the selected disks.")
                       % {"product": escape_markup(productName)})
        label = self.builder.get_object("no_space_desc_label")
        label.set_markup(label_text)
        label.connect("activate-link", self._modify_sw_link_clicked)

        self._set_free_space_labels(disk_free, fs_free)

        label_text = _("<b>You don't have enough space available to install "
                       "%(productName)s</b>, even if you used all of the free space "
                       "available on the selected disks.  You could add more "
                       "disks for additional space, "
                       "modify your software selection to install a smaller "
                       "version of <b>%(productName)s</b>, or quit the installer.") % \
                               {"productName": escape_markup(productName)}
        self.builder.get_object("no_space_options_label").set_markup(label_text)

        self._add_modify_watcher(label)

class StorageSpoke(NormalSpoke, StorageChecker):
    builderObjects = ["storageWindow", "backErrorDialog", "reclaimSpaceDialog"]
    mainWidgetName = "storageWindow"
    uiFile = "spokes/storage.glade"
    helpFile = "StorageSpoke.xml"

    category = SystemCategory

    # other candidates: computer-symbolic, folder-symbolic
    icon = "drive-harddisk-symbolic"
    title = CN_("GUI|Spoke", "INSTALLATION _DESTINATION")

    def __init__(self, *args, **kwargs):
        StorageChecker.__init__(self, min_ram=isys.MIN_GUI_RAM)
        NormalSpoke.__init__(self, *args, **kwargs)
        self.applyOnSkip = True

        self._ready = False
        self.autoPartType = None
        self.encrypted = False
        self.passphrase = ""
        self.selected_disks = self.data.ignoredisk.onlyuse[:]
        self.autopart_missing_passphrase = False
        self.disks_errors = []
        self.back_clicked = False

        # This list contains all possible disks that can be included in the install.
        # All types of advanced disks should be set up for us ahead of time, so
        # there should be no need to modify this list.
        self.disks = []

        if not flags.automatedInstall:
            # default to using autopart for interactive installs
            self.data.autopart.autopart = True

        self.autopart = self.data.autopart.autopart
        self.autoPartType = None
        self.clearPartType = CLEARPART_TYPE_NONE

        if arch.isS390() and (self.data.zerombr.zerombr or self.data.clearpart.cdl):
            # run dasdfmt on any unformatted or LDL DASDs automatically
            threadMgr.add(AnacondaThread(name=constants.THREAD_DASDFMT,
                            target=self.run_dasdfmt, args=(self.data.zerombr.zerombr, self.data.clearpart.cdl)))

        self._previous_autopart = False

        self._last_selected_disks = []

        self._grabObjects()

        self._local_disks_grid_height = 0

    def _grabObjects(self):
        self._customPart = self.builder.get_object("customRadioButton")

    def apply(self):
        applyDiskSelection(self.storage, self.data, self.selected_disks)

        for disk in self.disks:
            if self.storage.vstorage_cs:
                disk.use_vzdata = disk.use_slash

        self.data.autopart.autopart = self.autopart
        self.data.autopart.type = self.autoPartType
        self.data.autopart.encrypted = self.encrypted
        self.data.autopart.passphrase = self.passphrase

        if self.data.bootloader.bootDrive and \
           self.data.bootloader.bootDrive not in self.selected_disks:
            self.data.bootloader.bootDrive = ""
            self.storage.bootloader.reset()

        self.data.clearpart.initAll = True

        if not self.autopart_missing_passphrase:
            self.clearPartType = CLEARPART_TYPE_ALL
            self.data.clearpart.type = self.clearPartType

        self.storage.config.update(self.data)
        self.storage.autoPartType = self.data.autopart.type
        self.storage.encryptedAutoPart = self.data.autopart.encrypted
        self.storage.encryptionPassphrase = self.data.autopart.passphrase

        # If autopart is selected we want to remove whatever has been
        # created/scheduled to make room for autopart.
        # If custom is selected, we want to leave alone any storage layout the
        # user may have set up before now.
        self.storage.config.clearNonExistent = self.data.autopart.autopart

    @gtk_action_nowait
    def execute(self):
        # Spawn storage execution as a separate thread so there's no big delay
        # going back from this spoke to the hub while StorageChecker.run runs.
        # Yes, this means there's a thread spawning another thread.  Sorry.
        threadMgr.add(AnacondaThread(name=constants.THREAD_EXECUTE_STORAGE,
                                     target=self._doExecute))

        # Register iSCSI to kickstart data
        iscsi_devices = []
        # Find all selected disks and add all iscsi disks to iscsi_devices list
        for d in [d for d in getDisks(self.storage.devicetree) if d.name in self.selected_disks]:
            # Get parents of a multipath devices
            if isinstance(d, MultipathDevice):
                for parent_dev in d.parents:
                    if (isinstance(parent_dev, iScsiDiskDevice)
                        and not parent_dev.ibft
                        and not parent_dev.offload):
                        iscsi_devices.append(parent_dev)
            # Add no-ibft iScsiDiskDevice. IBFT disks are added automatically so there is
            # no need to have them in KS.
            elif isinstance(d, iScsiDiskDevice) and not d.ibft and not d.offload:
                iscsi_devices.append(d)

        if iscsi_devices:
            self.data.iscsiname.iscsiname = self.storage.iscsi.initiator
            # Remove the old iscsi data information and generate new one
            self.data.iscsi.iscsi = []
            for device in iscsi_devices:
                iscsi_data = self._create_iscsi_data(device)
                for saved_iscsi in self.data.iscsi.iscsi:
                    if (iscsi_data.ipaddr == saved_iscsi.ipaddr and
                        iscsi_data.target == saved_iscsi.target and
                        iscsi_data.port == saved_iscsi.port):
                        break
                else:
                    self.data.iscsi.iscsi.append(iscsi_data)

    def _doExecute(self):
        self._ready = False
        hubQ.send_not_ready(self.__class__.__name__)
        # on the off-chance dasdfmt is running, we can't proceed further
        threadMgr.wait(constants.THREAD_DASDFMT)
        hubQ.send_message(self.__class__.__name__, _("Saving storage configuration..."))
        threadMgr.wait(constants.THREAD_STORAGE)
        if flags.automatedInstall and self.data.autopart.encrypted and not self.data.autopart.passphrase:
            self.autopart_missing_passphrase = True
            StorageChecker.errors = [_("Passphrase for autopart encryption not specified.")]
            self._ready = True
            hubQ.send_ready(self.__class__.__name__, True)
            return
        try:
            doKickstartStorage(self.storage, self.data, self.instclass)
        except (StorageError, KickstartValueError) as e:
            log.error("storage configuration failed: %s", e)
            StorageChecker.errors = str(e).split("\n")
            hubQ.send_message(self.__class__.__name__, _("Failed to save storage configuration..."))
            self.data.bootloader.bootDrive = ""
            self.data.ignoredisk.drives = []
            self.data.ignoredisk.onlyuse = []
            self.storage.config.update(self.data)
            self.storage.reset()

            # HDDs preselect
            DoVZAutopartPreselect(self.instclass, self.storage, self.data).preselect_drives()

            self.disks = getDisks(self.storage.devicetree)
            self._update_disk_list()

            # now set ksdata back to the user's specified config
            applyDiskSelection(self.storage, self.data, self.selected_disks)
        except BootLoaderError as e:
            log.error("BootLoader setup failed: %s", e)
            StorageChecker.errors = str(e).split("\n")
            hubQ.send_message(self.__class__.__name__, _("Failed to save storage configuration..."))
            self.data.bootloader.bootDrive = ""
        else:
            if self.autopart or (flags.automatedInstall and (self.data.autopart.autopart or self.data.partition.seen)):
                # run() executes StorageChecker.checkStorage in a seperate thread
                self.run()
        finally:
            resetCustomStorageData(self.data)
            self._ready = True
            hubQ.send_ready(self.__class__.__name__, True)

    def _create_iscsi_data(self, device):
        from pyanaconda.kickstart import AnacondaKSHandler
        handler = AnacondaKSHandler()
        # pylint: disable=E1101
        iscsi_data = handler.IscsiData()
        dev_node = device.node
        iscsi_data.ipaddr = dev_node.address
        iscsi_data.target = dev_node.name
        iscsi_data.port = dev_node.port
        # Bind interface to target
        if self.storage.iscsi.ifaces:
            iscsi_data.iface = self.storage.iscsi.ifaces[dev_node.iface]

        auth = dev_node.getAuth()
        if auth:
            if auth.username and auth.password:
                iscsi_data.user = auth.username
                iscsi_data.password = auth.password
            if auth.reverse_username and auth.reverse_password:
                iscsi_data.user_in = auth.reverse_username
                iscsi_data.password_in = auth.reverse_password
        return iscsi_data

    @property
    def completed(self):
        retval = (threadMgr.get(constants.THREAD_EXECUTE_STORAGE) is None and
                  threadMgr.get(constants.THREAD_CHECK_STORAGE) is None and
                  self.storage.rootDevice is not None and
                  not self.errors)
        return retval

    @property
    def ready(self):
        # By default, the storage spoke is not ready.  We have to wait until
        # storageInitialize is done.
        return self._ready

    @property
    def showable(self):
        return not flags.dirInstall

    @property
    def status(self):
        """ A short string describing the current status of storage setup. """
        msg = _("No disks selected")

        if flags.automatedInstall and not self.storage.rootDevice:
            return msg
        elif threadMgr.get(constants.THREAD_DASDFMT):
            msg = _("Formatting DASDs")
        elif self.data.ignoredisk.onlyuse:
            msg = P_(("%d disk selected"),
                     ("%d disks selected"),
                     len(self.data.ignoredisk.onlyuse)) % len(self.data.ignoredisk.onlyuse)

            if self.errors:
                msg = _("Error checking storage configuration")
            elif self.warnings:
                msg = _("Invalid storage configuration")
            elif self.data.autopart.autopart:
                msg = _("Automatic partitioning selected")
            else:
                msg = _("Custom partitioning selected")

        return msg

    def refresh(self):
        self.disks = getDisks(self.storage.devicetree)

        self.back_clicked = False

        initialized = [ d for d in self.disks if self._is_disk_possible_to_select(d) ]

        if not initialized:
            # HDDs preselect (rescan)
            DoVZAutopartPreselect(self.instclass, self.storage, self.data).preselect_drives()

        # synchronize our local data store with the global ksdata
        disk_names = [d.name for d in self.disks]
        # don't put disks with hidden formats in selected_disks
        self.selected_disks = [d for d in self.data.ignoredisk.onlyuse
                                    if d in disk_names]

        # unhide previously hidden disks so that they don't look like being
        # empty (because of all child devices hidden)
        self._unhide_disks()

        self.autopart = self.data.autopart.autopart
        self.autoPartType = self.data.autopart.type
        if self.autoPartType is None:
            self.autoPartType = AUTOPART_TYPE_LVM
        self.encrypted = self.data.autopart.encrypted
        self.passphrase = self.data.autopart.passphrase

        self._previous_autopart = self.autopart

        if self._local_disks_grid_height:
            for i in range(0, self._local_disks_grid_height):
                self.local_disks_grid.remove_row(0)

        self._local_disks_grid_height = 0
        self._system_rb_group = None

        # Add disks to the grid and update the selections
        self._vz_add_header(self.local_disks_grid)

        for disk in filter(isLocalDisk, self.disks):
            # While technically local disks, zFCP devices are specialized
            # storage and should not be shown here.
            if disk.type is not "zfcp":
                self._vz_add_disk(disk, self.disks.index(disk), self.local_disks_grid)

        self.local_disks_grid.show_all()

        if not iutil.expert_mode() and self._local_disks_grid_height:
            for i in range(0, self._local_disks_grid_height):
                really_hide(self.local_disks_grid.get_child_at(COL_BOOT, i))

        self._customPart.set_active(not self.autopart)

        self._update_disk_list()
        self._update_summary()

        self._check_problems()

    def _check_problems(self):
        if self.errors:
            self.set_warning(_("Error checking storage configuration.  <a href=\"\">Click for details.</a>"))
            return True
        elif self.warnings:
            self.set_warning(_("Invalid storage configuration.  <a href=\"\">Click for details.</a>"))
            return True
        if minimal_hardware(self.disks, self.storage.vstorage_cs):
            self._autopartRadioButton.set_sensitive(False)
            self._customPart.set_active(True)
            self.set_warning(_("All available disks are too small to install Virtuozzo on them."))
            return True

        self._autopartRadioButton.set_sensitive(True)
        return False

    def _validate_datastore(self):
        vzdata = False
        csdata = False
        for disk in self.disks:
            if disk.use_vzdata:
                vzdata = True
            if disk.use_csdata:
                csdata = True

        if self.storage.vstorage_cs:
            return vzdata and csdata

        return vzdata

    def _validate_boot(self):
        return [d for d in self.disks if d.cap_boot and d.use_boot];

    def _set_title(self, label):
        bold_font = Pango.FontDescription()
        bold_font.set_weight(Pango.Weight.BOLD)
        label.modify_font(bold_font)

    def _set_note(self, label, text):
        label.set_markup("<span style=\"italic\" color=\"Gray\">" + text + "</span>")
        label.set_margin_top(6)
        label.set_margin_bottom(6)

    def _set_warning(self, label, text):
        label.set_markup("<span color=\"Red\">" + text + "</span>")

    def _on_back_error_ok_clicked(self, button, user_data=None):
        self._backErrorDialog.hide()

    def _on_reclaim_space_ok_clicked(self, button, user_data=None):
        self._delete_existing_part = True
        self._reclaimSpaceDialog.hide()

    def _on_reclaim_space_cancel_clicked(self, button, user_data=None):
        self._delete_existing_part = False
        self._reclaimSpaceDialog.hide()

    def initialize(self):
        NormalSpoke.initialize(self)

        self._backErrorDialog = self.builder.get_object("backErrorDialog")
        self._backErrorLabel = self.builder.get_object("backErrorLabel")
        self._reclaimSpaceDialog = self.builder.get_object("reclaimSpaceDialog")
        self._autopartRadioButton = self.builder.get_object("autopartRadioButton")

        self.local_disks_grid = self.builder.get_object("local_disks_grid")
        self.local_disks_grid.set_border_width(1)
        self.local_disks_grid.set_column_spacing(20)
        self.local_disks_grid.set_row_spacing(5)

        # Connect the viewport adjustments to the child widgets
        # See also https://bugzilla.gnome.org/show_bug.cgi?id=744721
        localViewport = self.builder.get_object("localViewport")
        localViewport.add(self.local_disks_grid)
        self.local_disks_grid.set_focus_vadjustment(localViewport.get_vadjustment())

        mainViewport = self.builder.get_object("storageViewport")
        mainBox = self.builder.get_object("storageMainBox")
        mainBox.set_focus_vadjustment(mainViewport.get_vadjustment())

        threadMgr.add(AnacondaThread(name=constants.THREAD_STORAGE_WATCHER,
                      target=self._initialize))

    def _vz_add_header(self, grid):
        # total
        self._local_disks_grid_height += 1

        # name
        name_label = Gtk.Label("Disk")
        self._set_title(name_label)
        name_label.set_halign(Gtk.Align.START)
        grid.attach(name_label, COL_DISK, 0, 1, 1)

        # space
        space_label = Gtk.Label("")
        space_label.set_size_request(50, -1)
        grid.attach(space_label, COL_SPACE, 0, 1, 1)

        # type
        type_label = Gtk.Label("Type")
        self._set_title(type_label)
        type_label.set_halign(Gtk.Align.START)
        grid.attach(type_label, COL_TYPE, 0, 1, 1)

        # size
        size_label = Gtk.Label("Size")
        self._set_title(size_label)
        size_label.set_halign(Gtk.Align.START)
        grid.attach(size_label, COL_SIZE, 0, 1, 1)

        # boot
        boot_label = Gtk.Label("Bootable")
        self._set_title(boot_label)
        boot_label.set_halign(Gtk.Align.START)
        grid.attach(boot_label, COL_BOOT, 0, 1, 1)

        # system
        system_label = Gtk.Label("System")
        self._set_title(system_label)
        system_label.set_halign(Gtk.Align.START)
        grid.attach(system_label, COL_SYSTEM, 0, 1, 1)

        # datastore
        datastore_label = Gtk.Label("Datastore")
        self._set_title(datastore_label)
        datastore_label.set_halign(Gtk.Align.START)
        grid.attach(datastore_label, COL_DATASTORE, 0, 1, 1)

        # cache
        cache_label = Gtk.Label("Cache")
        self._set_title(cache_label)
        cache_label.set_halign(Gtk.Align.START)
        grid.attach(cache_label, COL_CACHE, 0, 1, 1)

    def _vz_add_disk(self, disk, ndx, grid):
        top = self._local_disks_grid_height

        # total
        self._local_disks_grid_height += 1

        if isinstance(disk, MultipathDevice):
            desc = disk.wwid.split(":")
            description = ":".join(desc[0:3]) + "..." + ":".join(desc[-4:])
        elif isinstance(disk, ZFCPDiskDevice):
            # manually mangle the desc of a zFCP device to be multi-line since
            # it's so long it makes the disk selection screen look odd
            description = _("FCP device %(hba_id)s\nWWPN %(wwpn)s\nLUN %(lun)s") % \
                            {"hba_id": disk.hba_id, "wwpn": disk.wwpn, "lun": disk.fcp_lun}
        else:
            description = getDiskDescription(disk)

        no_space = "No space"

        # name
        name_label = Gtk.Label(disk.name + " / " + description)
        name_label.set_halign(Gtk.Align.START)
        name_label.set_property("tooltip-text", disk.serial)
        name_label.set_line_wrap(True)
        grid.attach(name_label, COL_DISK, top, 1, 1)
        
        # space
        space_label = Gtk.Label("")
        grid.attach(space_label, COL_SPACE, top, 1, 1)

        # type
        dtype = "HDD"
        if disk.is_ssd:
            dtype = "SSD"
        type_label = Gtk.Label(dtype)
        type_label.set_halign(Gtk.Align.START)
        grid.attach(type_label, COL_TYPE, top, 1, 1)

        # size
        # free = self.storage.getFreeSpace(disks=[disk])[disk.name][0]
        avail_size = Size(disk.size).__str__()
        if self._cap_datastore(disk) and disk.cap_slash:
            size_label = Gtk.Label(avail_size)
        else:
            size_label = Gtk.Label(avail_size)
            # self._set_warning(size_label, avail_size)

        size_label.set_halign(Gtk.Align.START)
        grid.attach(size_label, COL_SIZE, top, 1, 1)

        # boot
        if disk.cap_boot:
            boot_cb = Gtk.CheckButton()
            boot_cb.connect("toggled", self._on_boot_disk_toggled, ndx, top)
            boot_cb.set_sensitive(True)
            boot_cb.set_active(disk.use_boot)
            boot_cb.set_halign(Gtk.Align.START)
            boot_cb.set_property("tooltip-text", "Add this disk to the boot menu")
        else:
            boot_cb = Gtk.Label()
            self._set_note(boot_cb, "N/A")
            boot_cb.set_halign(Gtk.Align.START)
            boot_cb.set_property("tooltip-text", "This disk cannot be added to the boot menu")

        grid.attach(boot_cb, COL_BOOT, top, 1, 1)

        # datastore
        if self._cap_datastore(disk):
            datastore_cb = Gtk.CheckButton()
            datastore_cb.connect("toggled", self._on_datastore_disk_toggled, ndx, top)
            datastore_cb.set_sensitive(self._cap_datastore(disk))
            datastore_cb.set_halign(Gtk.Align.START)
            datastore_cb.set_property("tooltip-text", "Use this disk for datastore")
        else:
            datastore_cb = Gtk.Label()
            self._set_note(datastore_cb, no_space)
            datastore_cb.set_halign(Gtk.Align.START)
            datastore_cb.set_property("tooltip-text", "There is not enough space on the disk to set up a datastore")

        grid.attach(datastore_cb, COL_DATASTORE, top, 1, 1)

        # system
        if disk.cap_slash:
            if self._system_rb_group:
                system_rb = Gtk.RadioButton.new_from_widget(self._system_rb_group)
            else:
                system_rb = Gtk.RadioButton(None)
                self._system_rb_group = system_rb

            system_rb.connect("toggled", self._on_system_disk_toggled, ndx, top)
            system_rb.set_sensitive(disk.cap_slash)
            system_rb.set_halign(Gtk.Align.START)
            system_rb.set_property("tooltip-text", "Use this disk for system")
        else:
            system_rb = Gtk.Label()
            self._set_note(system_rb, no_space)
            system_rb.set_halign(Gtk.Align.START)
            system_rb.set_property("tooltip-text", "There is not enough space on the disk to set up a system")

        grid.attach(system_rb, COL_SYSTEM, top, 1, 1)

        # cache
        if disk.is_ssd:
            cache_cb = Gtk.CheckButton()
            cache_cb.connect("toggled", self._on_cache_disk_toggled, ndx, top)
            cache_cb.set_sensitive(True)
            cache_cb.set_halign(Gtk.Align.START)
            cache_cb.set_property("tooltip-text", "Use this disk for cache")
        else:
            cache_cb = Gtk.Label()
            self._set_note(cache_cb, "N/A")
            cache_cb.set_halign(Gtk.Align.START)
            cache_cb.set_property("tooltip-text", "This disk cannot be used for cache (non-SSD)")

        grid.attach(cache_cb, COL_CACHE, top, 1, 1)

        # activate system & datastore
        if self._cap_datastore(disk):
            datastore_cb.set_active(self._used_datastore(disk))

        if disk.cap_slash:
            system_rb.set_active(disk.use_slash)
        
        if disk.is_ssd:
            cache_cb.set_active(disk.use_ssd)

    def _on_system_disk_toggled(self, button, ndx, top):
        state = button.get_active()

        self._set_boot_state(ndx, top)

        if self.disks[ndx].is_ssd:
            cache_cb = self.local_disks_grid.get_child_at(COL_CACHE, top)
            if not state:
                if self._cap_datastore(self.disks[ndx]):
                    datastore_cb = self.local_disks_grid.get_child_at(COL_DATASTORE, top)
                    cache_cb.set_sensitive(not datastore_cb.get_active())
                    return
                cache_cb.set_sensitive(True)
            else:
                cache_cb.set_sensitive(False)

        if not state:
            return

        for disk in self.disks:
            disk.use_slash = False

        self.disks[ndx].use_slash = True

        # deselect Datastore
        if not self._single_system_disk() and self._cap_datastore(self.disks[ndx]):
            self.local_disks_grid.get_child_at(COL_DATASTORE, top).set_active(False)

        self._update_disk_list()
        self._update_summary()

    def _single_system_disk(self):
        count = 0
        for disk in self.disks:
            if disk.cap_slash:
                count += 1

        return count == 1

    def _on_boot_disk_toggled(self, button, ndx, top):
        self.disks[ndx].use_boot = button.get_active()
        self._update_disk_list()
        self._update_summary()

    def _on_datastore_disk_toggled(self, button, ndx, top):
        state = button.get_active()

        self._set_boot_state(ndx, top)

        if self.disks[ndx].is_ssd:
            cache_cb = self.local_disks_grid.get_child_at(COL_CACHE, top)
            if state:
                cache_cb.set_sensitive(False)
            else:
                system_rb = self.local_disks_grid.get_child_at(COL_SYSTEM, top)
                cache_cb.set_sensitive(not system_rb.get_active())

        self._select_datastore(self.disks[ndx], state)

        self._update_disk_list()
        self._update_summary()

    def _on_cache_disk_toggled(self, button, ndx, top):
        state = button.get_active()

        self._set_boot_state(ndx, top)

        self.disks[ndx].use_ssd = state

        if self.disks[ndx].cap_slash:
            system_rb = self.local_disks_grid.get_child_at(COL_SYSTEM, top)
            system_rb.set_sensitive(not state)

        if self._cap_datastore(self.disks[ndx]):
            datastore_cb = self.local_disks_grid.get_child_at(COL_DATASTORE, top)
            datastore_cb.set_sensitive(not state)

        self._update_disk_list()
        self._update_summary()

    def _set_boot_state(self, ndx, top):
        if not self.disks[ndx].cap_boot:
            return

        # check if any of the checkboxes is enabled
        any_active = False

        if self.disks[ndx].cap_slash:
            system_cb = self.local_disks_grid.get_child_at(COL_SYSTEM, top)
            any_active = system_cb.get_active()

        if not any_active and self._cap_datastore(self.disks[ndx]):
            datastore_cb = self.local_disks_grid.get_child_at(COL_DATASTORE, top)
            any_active = datastore_cb.get_active()

        if not any_active and self.disks[ndx].is_ssd:
            cache_cb = self.local_disks_grid.get_child_at(COL_CACHE, top)
            any_active = cache_cb.get_active()

        boot_cb = self.local_disks_grid.get_child_at(COL_BOOT, top)
        boot_cb.set_active(any_active)
        boot_cb.set_sensitive(not any_active)

    def _cap_datastore(self, disk):
        if self.storage.vstorage_cs:
            return disk.cap_csdata

        return disk.cap_vzdata

    def _used_datastore(self, disk):
        if self.storage.vstorage_cs:
            return disk.use_csdata

        return disk.use_vzdata

    def _select_datastore(self, disk, state):
        if self.storage.vstorage_cs:
            if disk.cap_csdata:
                disk.use_csdata = state
        else:
            if disk.cap_vzdata:
                disk.use_vzdata = state

    def _initialize(self):
        hubQ.send_message(self.__class__.__name__, _("Probing storage..."))

        threadMgr.wait(constants.THREAD_STORAGE)
        threadMgr.wait(constants.THREAD_CUSTOM_STORAGE_INIT)

        # HDDs preselect
        DoVZAutopartPreselect(self.instclass, self.storage, self.data).preselect_drives()

        self.disks = getDisks(self.storage.devicetree)
        self._update_disk_list()

        applyDiskSelection(self.storage, self.data, self.selected_disks)

        self._ready = True
        hubQ.send_ready(self.__class__.__name__, False)

    def _update_summary(self):
        """ Update the summary based on the UI. """
        count = 0
        capacity = Size(0)
        free = Size(0)

        # pass in our disk list so hidden disks' free space is available
        free_space = self.storage.getFreeSpace(disks=self.disks)
        selected = [d for d in self.disks if d.name in self.selected_disks]

        for disk in selected:
            capacity += disk.size
            free += free_space[disk.name][0]
            count += 1

        anySelected = count > 0

        summary = (P_("%(count)d disk selected; %(capacity)s capacity; %(free)s free",
                      "%(count)d disks selected; %(capacity)s capacity; %(free)s free",
                      count) % {"count" : count,
                                "capacity" : capacity,
                                "free" : free})
        summary_label = self.builder.get_object("summary_label")
        summary_label.set_text(summary)
        summary_label.set_sensitive(anySelected)

        # only show the "we won't touch your other disks" labels and summary button when
        # some disks are selected
        self.builder.get_object("local_untouched_label_revealer").set_reveal_child(anySelected)
        self.builder.get_object("other_options_label").set_sensitive(anySelected)
        self.builder.get_object("other_options_grid").set_sensitive(anySelected)

        if len(self.disks) == 0:
            self.set_warning(_("No disks detected.  Please shut down the computer, connect at least one disk, and restart to complete installation."))
        elif not anySelected:
            # There may be an underlying reason that no disks were selected, give them priority.
            if not self._check_problems():
                self.set_warning(_("No disks selected; please select at least one disk to install to."))
        else:
            self.clear_info()

    def _select_disk(self, disk_name, select):
        if select:
            if disk_name not in self.selected_disks:
                self.selected_disks.append(disk_name)
        else:
            if disk_name in self.selected_disks:
                self.selected_disks.remove(disk_name)

    def _update_disk_list(self):
        for disk in self.disks:
            self._select_disk(disk.name, self._is_disk_selected(disk))

    def _is_disk_selected(self, disk):
        if minimal_hardware(self.disks, self.storage.vstorage_cs):
            return True
        if iutil.expert_mode() and disk.use_boot:
            return True
        return disk.use_slash or self._used_datastore(disk) or disk.use_ssd

    def _is_disk_possible_to_select(self, disk):
        return disk.cap_slash or disk.cap_csdata or disk.cap_vzdata

    def run_dasdfmt(self, doformat, cdl):
        """
        Though the same function exists in pyanaconda.ui.gui.spokes.lib.dasdfmt,
        this instance doesn't include any of the UI pieces and should only
        really be getting called on ks installations that specify zerombr.
        """
        # wait for the initial storage thread to complete before taking any new
        # actions on storage devices
        threadMgr.wait(constants.THREAD_STORAGE)
        updated = False

        if doformat:
            unformatted = make_unformatted_dasd_list([d.name for d in getDisks(self.storage.devicetree)])
            if unformatted:
                hubQ.send_message(self.__class__.__name__, _("Formatting DASDs"))
                for disk in unformatted:
                    try:
                        format_dasd(disk)
                        # call removeChildren function instead of simply
                        # removeDevice since the disk may have children in
                        # devicetree, e.g. /dev/dasdc may have /dev/dasdc1
                        remove = self.storage.devicetree.getDeviceByName(disk)
                        self.storage.devicetree._removeChildrenFromTree(remove)
                    except DasdFormatError as err:
                        # Log errors if formatting fails, but don't halt the installer
                        log.error(str(err))
                        continue

                    updated = True

        if cdl:
            ldldasds = [d for d in self.storage.devicetree.dasd if is_ldl_dasd(d.name)]
            if ldldasds:
                hubQ.send_message(self.__class__.__name__, _("Formatting DASDs"))
                for disk in ldldasds:
                    try:
                        format_dasd(disk.name)
                        # call removeChildren function instead of simply
                        # removeDevice since the disk may have children in
                        # devicetree, e.g. /dev/dasdc may have /dev/dasdc1
                        self.storage.devicetree._removeChildrenFromTree(disk)
                    except DasdFormatError as err:
                        # Log errors if formatting fails, but don't halt the installer
                        log.error(str(err))
                        continue

                    updated = True

        if updated:
            self.storage.devicetree.populate()

    def run_lightbox_dialog(self, dialog):
        with self.main_window.enlightbox(dialog.window):
            rc = dialog.run()

        return rc

    def _check_encrypted(self):
        # even if they're not doing autopart, setting autopart.encrypted
        # establishes a default of encrypting new devices
        if not self.encrypted:
            return True

        dialog = PassphraseDialog(self.data)
        rc = self.run_lightbox_dialog(dialog)
        if rc != 1:
            return False

        self.passphrase = dialog.passphrase

        for device in self.storage.devices:
            if device.format.type == "luks" and not device.format.exists:
                if not device.format.hasKey:
                    device.format.passphrase = self.passphrase

        return True

    def _hide_disks(self):
        for disk in self.disks:
            if disk.name not in self.selected_disks and \
               disk in self.storage.devices:
                self.storage.devicetree.hide(disk)


    def _unhide_disks(self):
        for disk in self.disks:
            if disk.name not in self.selected_disks and \
               disk.name not in self._last_selected_disks:
                self.storage.devicetree.unhide(disk)

    def _remove_nonexistant_partitions(self):
        for partition in self.storage.partitions[:]:
            # check if it's been removed in a previous iteration
            if not partition.exists and \
               partition in self.storage.partitions:
                self.storage.recursiveRemove(partition)

    def on_back_clicked(self, button):
        # We can't exit early if it looks like nothing has changed because the
        # user might want to change settings presented in the dialogs shown from
        # within this method.

        # Do not enter this method multiple times if user clicking multiple times
        # on back button
        if self.back_clicked:
            return
        else:
            self.back_clicked = True

        if not self._validate_datastore() and not \
                minimal_hardware(self.disks, self.storage.vstorage_cs):
            self._backErrorLabel.set_text("At least one disk for datastore need to be selected.")
            with self.main_window.enlightbox(self._backErrorDialog):
                self._backErrorDialog.run()
            self.back_clicked = False
            return
        elif not self._validate_boot():
            self._backErrorLabel.set_text("Select at least one disk to be made bootable.")
            with self.main_window.enlightbox(self._backErrorDialog):
                self._backErrorDialog.run()
            self.back_clicked = False
            return

        if self.autopart_missing_passphrase:
            self._check_encrypted()
            NormalSpoke.on_back_clicked(self, button)
            return

        # No disks selected?  The user wants to back out of the storage spoke.
        if not self.selected_disks:
            NormalSpoke.on_back_clicked(self, button)
            return

        disk_selection_changed = False
        if self._last_selected_disks:
            disk_selection_changed = (self._last_selected_disks != set(self.selected_disks))

        # remember the disk selection for future decisions
        self._last_selected_disks = set(self.selected_disks)

        autopart = not self._customPart.get_active()

        # hide disks as requested
        self._hide_disks()

        # make sure no containers were split up by the user's disk selection
        self.clear_info()

        # if there are some disk selection errors we don't let user to leave the
        # spoke, so these errors don't have to go to self.errors
        self.disks_errors = checkDiskSelection(self.storage, self.selected_disks)
        if self.disks_errors:
            # The disk selection has to make sense before we can proceed.
            self.set_error(_("There was a problem with your disk selection. "
                             "Click here for details."))
            self._unhide_disks()
            self.back_clicked = False
            return

        # hide/unhide disks as requested
        for disk in self.disks:
            if disk.name not in self.selected_disks and \
               disk in self.storage.devices:
                self.storage.devicetree.hide(disk)
            elif disk.name in self.selected_disks and \
                 disk not in self.storage.devices:
                self.storage.devicetree.unhide(disk)

        if arch.isS390():
            # check for unformatted DASDs
            unformatted = make_unformatted_dasd_list(self.selected_disks)
            # check for ldl dasds
            ldldasds = [d for d in self.selected_disks if is_ldl_dasd(d)]

            # combine it all in one list and format
            dasds = list(set(unformatted + ldldasds))
            if len(dasds) > 0:
                # need to get a list of DASDDevice objects, so....
                dasdobjs = [self.storage.devicetree.getDeviceByName(d) for d in dasds]

                # We want to apply current selection before running dasdfmt to
                # prevent this information from being lost afterward
                applyDiskSelection(self.storage, self.data, self.selected_disks)
                dialog = DasdFormatDialog(self.data, self.storage, dasdobjs)
                ignoreEscape(dialog.window)
                rc = self.run_lightbox_dialog(dialog)
                if rc == 1:
                    # User hit OK on the dialog
                    self.refresh()
                elif rc == 2:
                    # User clicked uri to return to hub.
                    NormalSpoke.on_back_clicked(self, button)
                    return
                elif rc != 2:
                    # User either hit cancel on the dialog or closed it via escape,
                    # there was no formatting done.
                    # NOTE: rc == 2 means the user clicked on the link that takes t
                    # back to the hub.
                    return

        disks = [d for d in self.disks if d.name in self.selected_disks]
        disks_size = sum((d.size for d in disks), Size(0))

        # Figure out if the existing disk labels will work on this platform
        # you need to have at least one of the platform's labels in order for
        # any of the free space to be useful.
        disk_labels = set(disk.format.labelType for disk in disks
                              if hasattr(disk.format, "labelType"))
        platform_labels = set(platform.diskLabelTypes)
        if disk_labels and platform_labels.isdisjoint(disk_labels):
            disk_free = 0
            fs_free = 0
            log.debug("Need disklabel: %s have: %s", ", ".join(platform_labels),
                                                     ", ".join(disk_labels))
        else:
            free_space = self.storage.getFreeSpace(disks=disks,
                                                   clearPartType=CLEARPART_TYPE_NONE)
            disk_free = sum(f[0] for f in free_space.itervalues())
            fs_free = sum(f[1] for f in free_space.itervalues())

        required_space = self.payload.spaceRequired
        auto_swap = sum((r.size for r in self.storage.autoPartitionRequests
                                if r.fstype == "swap"), Size(0))
        if self.autopart and auto_swap == Size(0):
            # autopartitioning requested, but not applied yet (=> no auto swap
            # requests), ask user for enough space to fit in the suggested swap
            auto_swap = swap_lib.swapSuggestion()

        log.debug("disk free: %s  fs free: %s  sw needs: %s  auto swap: %s",
                  disk_free, fs_free, required_space, auto_swap)

        rc = RESPONSE_RECLAIM
        dialog = True

        if not dialog:
            # Plenty of room - there's no need to pop up a dialog, so just send
            # the user to wherever they asked to go.  That's either the custom
            # spoke or the hub.
            #    - OR -
            # Not enough room, but the user checked the reclaim button.

            self.encrypted = False

            if self._customPart.get_active():
                self.autopart = False
                self.skipTo = "CustomPartitioningSpoke"
            else:
                self.autopart = True

                # We might first need to ask about an encryption passphrase.
                if not self._check_encrypted():
                    return
        elif rc == RESPONSE_CANCEL:
            # A cancel button was clicked on one of the dialogs.  Stay on this
            # spoke.  Generally, this is because the user wants to add more disks.
            return
        elif rc == RESPONSE_MODIFY_SW:
            # The "Fedora software selection" link was clicked on one of the
            # dialogs.  Send the user to the software spoke.
            self.skipTo = "SoftwareSelectionSpoke"
        elif rc == RESPONSE_RECLAIM:
            # Not enough space, but the user can make enough if they do some
            # work and free up space.
            self.encrypted = False

            if not self._check_encrypted():
                self.back_clicked = False
                return

            if not self._show_resize_dialog(disks):
                # User pressed cancel on the reclaim dialog, so don't leave
                # the storage spoke.
                self.back_clicked = False
                return

            self.apply()

            # And then go to the custom partitioning spoke if they chose to
            # do so.
            if self._customPart.get_active():
                self.autopart = False
                self.skipTo = "CustomPartitioningSpoke"
            else:
                self.autopart = True
        elif rc == RESPONSE_QUIT:
            # Not enough space, and the user can't do anything about it so
            # they chose to quit.
            raise SystemExit("user-selected exit")
        else:
            # I don't know how we'd get here, but might as well have a
            # catch-all.  Just stay on this spoke.
            return

        if self.autopart:
            refreshAutoSwapSize(self.storage)
        self.applyOnSkip = True
        NormalSpoke.on_back_clicked(self, button)

    def _show_resize_dialog(self, disks):
        with self.main_window.enlightbox(self._reclaimSpaceDialog):
            self._reclaimSpaceDialog.run()
        return self._delete_existing_part

    def on_info_bar_clicked(self, *args):
        if self.disks_errors:
            label = _("The following errors were encountered when checking your disk "
                      "selection. You can modify your selection or quit the "
                      "installer.")

            dialog = DetailedErrorDialog(self.data, buttons=[
                    C_("GUI|Storage|Error Dialog", "_Quit"),
                    C_("GUI|Storage|Error Dialog", "_Modify Disk Selection")],
                label=label)
            with self.main_window.enlightbox(dialog.window):
                errors = "\n".join(self.disks_errors)
                dialog.refresh(errors)
                rc = dialog.run()

            dialog.window.destroy()

            if rc == 0:
                # Quit.
                iutil.ipmi_report(constants.IPMI_ABORTED)
                sys.exit(0)

        elif self.errors:
            label = _("The following errors were encountered when checking your storage "
                      "configuration.  You can modify your storage layout or quit the "
                      "installer.")

            dialog = DetailedErrorDialog(self.data, buttons=[
                    C_("GUI|Storage|Error Dialog", "_Quit"),
                    C_("GUI|Storage|Error Dialog", "_Modify Storage Layout")],
                label=label)
            with self.main_window.enlightbox(dialog.window):
                errors = "\n".join(self.errors)
                dialog.refresh(errors)
                rc = dialog.run()

            dialog.window.destroy()

            if rc == 0:
                # Quit.
                sys.exit(0)
                iutil.ipmi_report(constants.IPMI_ABORTED)
        elif self.warnings:
            label = _("The following warnings were encountered when checking your storage "
                      "configuration.  These are not fatal, but you may wish to make "
                      "changes to your storage layout.")

            dialog = DetailedErrorDialog(self.data, buttons=[_("_OK")], label=label)
            with self.main_window.enlightbox(dialog.window):
                warnings = "\n".join(self.warnings)
                dialog.refresh(warnings)
                rc = dialog.run()

            dialog.window.destroy()

    # This callback is for the button that has anaconda go back and rescan the
    # disks to pick up whatever changes the user made outside our control.
    def on_refresh_clicked(self, *args):
        dialog = RefreshDialog(self.data, self.storage)
        ignoreEscape(dialog.window)
        with self.main_window.enlightbox(dialog.window):
            rc = dialog.run()
            dialog.window.destroy()

        if rc == 1:
            # User hit OK on the dialog, indicating they stayed on the dialog
            # until rescanning completed.
            on_disk_storage.dispose_snapshot()
            self.refresh()
            return
        elif rc != 2:
            # User either hit cancel on the dialog or closed it via escape, so
            # there was no rescanning done.
            # NOTE: rc == 2 means the user clicked on the link that takes them
            # back to the hub.
            return

        on_disk_storage.dispose_snapshot()

        # Can't use this spoke's on_back_clicked method as that will try to
        # save the right hand side, which is no longer valid.  The user must
        # go back and select their disks all over again since whatever they
        # did on the shell could have changed what disks are available.
        NormalSpoke.on_back_clicked(self, None)
